import unittest
from unittest.mock import patch

from tmc import points, reflect
from tmc.utils import load, load_module, reload_module, get_stdout, check_source
from functools import reduce
import os
import os.path
import textwrap
import inspect, re
from random import choice, randint, shuffle

exercise = 'src.best_exam_result'

def source_rows(funktio: callable):
    src = inspect.getsource(funktio)
    lines = [line.strip() for line in re.split('\\n|;', src) 
        if len(line.strip()) > 0 and not line.strip().startswith("#")]
    return len(lines)

@points('11.best_exam_result')
class BestExamResultTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with patch('builtins.input', side_effect=[AssertionError("Asking input from the user was not expected.")]):
           cls.module = load_module(exercise, 'en')

    def test_0a_main_program_ok(self):
        ok, line = check_source(self.module)
        message = """The code for testing the functions should be placed inside
if __name__ == "__main__":
block. The following row should be moved:
"""
        self.assertTrue(ok, message+line)
    
    def test_1_function_exists(self):
        try:
            from src.best_exam_result import best_results
        except Exception as e:
            self.fail(f'Your program should contain function named best_results.')

    def test_1b_class_exists(self):
        try:
            from src.best_exam_result import ExamResult
        except Exception as e:
            self.fail(f'Your program should contain class named ExamResult - ' + 
                'Make sure, that the class is defined.')

    def test_2_type_of_return_value(self):
        try:
            from src.best_exam_result import best_results, ExamResult
            val = best_results([ExamResult("Peter",1,2,3)])
        except Exception as e:
            self.assertTrue(False, f'Function threw an error when it was called as follows:\n' + 
                'best_results([ExamResult("Peter",1,2,3)]):\n{e}')
        taip = str(type(val)).replace("<class '","").replace("'>","")
        self.assertTrue(type(val) == list, f"Function best_results is expected to return a value which is of type list," +  
            f" now it returns a value {val} which is of type {taip}\n when it is called with the parameter" +  
            '\nbest_results([ExamResult("Peter",1,2,3)])')
        

    def test_3_length_of_function(self):
        from src.best_exam_result import best_results
        lines = source_rows(best_results)
        max_lines = 2
        self.assertTrue(lines <= max_lines, f'Function best_results must have at most'+ 
            f' {max_lines} rows in this exercise.\n' +
            f'The function now has a total of {lines} rows (excluding empty rows and comments).')

    def test_4_test_with_values_1(self):
        test_cases = [("Pasi",3,1,2), ("Kimmo",4,5,3), ("Arto",2,3,4)]
        from src.best_exam_result import best_results, ExamResult
        input_values = [ExamResult(x[0],x[1],x[2],x[3]) for x in test_cases]
        corr = [max(x[1:]) for x in test_cases]
        val = best_results(input_values)
        test_f = ", ".join([f'ExamResult{x}' for x in test_cases])

        self.assertEqual(val, corr, f'The function is expected to return the following list:\n{corr}\n' +
            f'when it is called with the parameter {test_f}\nnow the function returns\n' + 
            f'{val}')

    def test_5_test_with_values_2(self):
        test_cases = [("Leenu",4,1,2), ("Linnu",1,2,5), ("Tiina",4,4,5), ("Matti",5,1,1), ("Keijo",2,4,2)]
        from src.best_exam_result import best_results, ExamResult
        input_values = [ExamResult(x[0],x[1],x[2],x[3]) for x in test_cases]
        corr = [max(x[1:]) for x in test_cases]
        val = best_results(input_values)
        test_f = ", ".join([f'ExamResult{x}' for x in test_cases])

        self.assertEqual(val, corr, f'The function is expected to return the following list:\n{corr}\n' +
            f'when it is called with the parameter {test_f}\nnow the function returns\n' + 
            f'{val}')

if __name__ == '__main__':
    unittest.main()
